#include <stdio.h>
#include <stdlib.h>
#include "lists.h"

void
list_init(struct list* l)
{
	l->head.next=&l->head;
	l->head.prev=&l->head;
}

int
list_is_empty(struct list* l)
{
	return (&l->head==l->head.next);
}

/* return first element of the list or NULL if empty */
void*
list_first(struct list* l)
{
	return (&l->head!=l->head.next?l->head.next:NULL);	
}

void*
list_last(struct list* l)
{
	return (&l->head!=l->head.prev?l->head.prev:NULL);	
}

void*
list_prev(struct list* l, struct node* n)
{
	return (n->prev!=&l->head?n->prev:NULL);
}

void*
list_next(struct list* l, struct node* n)
{
	return (n->next!=&l->head?n->next:NULL);
}

void
list_add_first(struct list* l, struct node* new)
{
	new->prev=&l->head;
	new->next=l->head.next;
	l->head.next->prev=new;
	l->head.next=new;
}

void
list_add_last(struct list* l, struct node* new)
{
	new->prev=l->head.prev;
	new->next=&l->head;
	l->head.prev->next=new;
	l->head.prev=new;
}

void
list_add_after(struct node* orig, struct node* new)
{
	new->prev=orig;
	new->next=orig->next;
	orig->next->prev=new;
	orig->next=new;
}

void
list_add_before(struct node* orig, struct node* new)
{
	new->prev=orig->prev;
	new->next=orig;
	orig->prev->next=new;
	orig->prev=new;
}

void
list_del_node(struct node* n)
{
	n->prev->next=n->next;
	n->next->prev=n->prev;
	n->prev=n->next=NULL;
}

void*
list_del_first(struct list* l)
{
	struct node* del;

	if (list_is_empty(l)) 
		return NULL;

	del=l->head.next;
	list_del_node(del);
	
	return del;
}

void*
list_del_last(struct list* l)
{
	struct node* del;

	if (list_is_empty(l)) 
		return NULL;

	del=l->head.prev;
	list_del_node(del);
	
	return del;
}

void
list_cat(struct list* dst, struct list* src) 
{
	struct node* dst_last = dst->head.prev;
	struct node* src_first = src->head.next;
	struct node* src_last = src->head.prev;

	dst_last->next = src_first;
	src_first->prev = dst_last;
	src_last->next = &dst->head;
	dst->head.prev = src_last;
}
